#!/usr/bin/env python3

__author__ = 'Stuart Longland VK4MSL'
__email__ = 'me@vk4msl.id.au'
__license__ = 'GPL-2.0-or-later'
__copyright__ = 'Copyright 2021, Stuart Longland (and contributors)'
__version__ = '0.0.11.post0'
